--[[
    Title: DKM, Fail Stat module
    Author: Sidh@Tarren Mill
    License: GPL
]]--

if select(2, UnitClass("player")) ~= "DEATHKNIGHT" then
   return
end

local L = LibStub("AceLocale-3.0"):GetLocale("DKM")

assert(DKM, L["Unable to locate DK Addon Object."])
assert(ModuleBase, L["Unable to locate ModuleBase object."])

local FailStat = DKM:NewModule("FailStat", ModuleBase, "AceConsole-3.0", "AceEvent-3.0")

-- Represents current state of each rune (true = ready, false = not ready)
FailStat.runesCurrentState = {
    [1] = true,
    [2] = true,
    [3] = true,
    [4] = true,
    [5] = true,
    [6] = true
}

-- Represents time from GetTime() when BOTH runes of the same type became available (eg. we are wasting refresh time)
FailStat.runeTypeReadyTimestamp = {
    [1] = nil,
    [2] = nil,
    [3] = nil
}

-- Represents accumulated wasted refresh time for runes of each type
FailStat.runesAccumulatedStats = {
    [1] = 0,
    [2] = 0,
    [3] = 0
}

local runeIDtoRuneType = {
    [1] = 1,
    [2] = 1,
    [3] = 2,
    [4] = 2,
    [5] = 3,
    [6] = 3
}

local runeIDPairs = {
    [1] = 2,
    [2] = 1,
    [3] = 4,
    [4] = 3,
    [5] = 6,
    [6] = 5
}

local runeTypeName = {
    [1] = 'Blood',
    [2] = 'Unholy',
    [3] = 'Frost'
}

function FailStat:OnInitialize()
    DKM.options.args.failstat = {
        name = "General",
        type = "group",
        desc = "General options",
        args = {
            logRuneStatistics = {
                type = "toggle",
                width = "full",
                name = "Log rune statistics",
                desc = "Logs waste of rune refresh time.",
                handler = FailStat,
                get = "IsLogRuneStatistics",
                set = "SetLogRuneStatistics",
            }
        }
    }
    
    self.db = DKM.db.profile.failstat
    self.options = DKM.options.args.failstat
    self:InitBaseOptions(L["Enables "..self:GetName().." module."])
    DKM.optionsFrameFailStat = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DKM", "Statistics", "DKM", "failstat")
    
    self:SetEnabledState(false)
end

function FailStat:OnEnable()    
    self:UpdateLogRuneStatistics()
    
    DKM:Print(L[self:GetName().." module enabled."])
end

function FailStat:OnDisable()
    self:UnregisterEvents()
    
    DKM:Print(L[self:GetName().." module disabled."])
end

function FailStat:RegisterEvents()
    self:RegisterEvent("PLAYER_REGEN_ENABLED")
    self:RegisterEvent("PLAYER_REGEN_DISABLED")
    self:RegisterEvent("RUNE_POWER_UPDATE")
end

function FailStat:UnregisterEvents()
    self:UnregisterEvent("PLAYER_REGEN_ENABLED")
    self:UnregisterEvent("PLAYER_REGEN_DISABLED")
    self:UnregisterEvent("RUNE_POWER_UPDATE")
end

function FailStat:PLAYER_REGEN_ENABLED()
    -- Add up current rune states
    for rune_type = 1,3 do
        if self:AreBothRunesReady( rune_type ) then
            self.runesAccumulatedStats[rune_type] = self.runesAccumulatedStats[rune_type] + ( GetTime() - self.runeTypeReadyTimestamp[ rune_type ] )
        end
    end
    
    DKM:Print('Combat over.')
    DKM:Print('Rune refresh time statistics:')
    for rune_type = 1,3 do
        DKM:Print('Wasted '..( ceil( self.runesAccumulatedStats[rune_type] * 10 ) / 10 )..' seconds for '..runeTypeName[rune_type]..' runes.')
    end
end

function FailStat:PLAYER_REGEN_DISABLED()    
    DKM:Print('Combat started. Logging statistics...')

    -- Reset current rune states
    for rune_id = 1,6 do
        self.runesCurrentState[rune_id] = self:IsRuneReady( rune_id )
    end
    
    -- Update rune types data
    for rune_type = 1,3 do
        -- Reset current rune type timestamp
        if self:AreBothRunesReady( rune_type ) then
            self.runeTypeReadyTimestamp[ rune_type ] = GetTime()
        else
            self.runeTypeReadyTimestamp[ rune_type ] = nil
        end
        
        -- Reset accumulated stat
        self.runesAccumulatedStats[rune_type] = 0
    end
end

function FailStat:RUNE_POWER_UPDATE(event, ...)
    if not self:IsInCombat() then
        return
    end
    
    local rune_id = ...
    local runeReady = self:IsRuneReady( rune_id )
    
    if self.runesCurrentState[rune_id] == runeReady then
        return
    end
    
    self.runesCurrentState[rune_id] = runeReady
    local rune_type = runeIDtoRuneType[rune_id]
    if self:AreBothRunesReady( rune_type ) then
        self.runeTypeReadyTimestamp[ rune_type ] = GetTime()
    else
        if not runeReady and self.runesCurrentState[ runeIDPairs[ rune_id ] ] then
            self.runesAccumulatedStats[rune_type] = self.runesAccumulatedStats[rune_type] + ( GetTime() - self.runeTypeReadyTimestamp[ rune_type ] )
        end
        self.runeTypeReadyTimestamp[ rune_type ] = nil
    end
end

function FailStat:AreBothRunesReady( rune_type )
    local rune_id = rune_type * 2
    return self.runesCurrentState[rune_id] and self.runesCurrentState[rune_id - 1]
end

function FailStat:IsInCombat()
    return InCombatLockdown()
end

function FailStat:IsRuneReady( rune_id )
    local start, duration, runeReady = GetRuneCooldown( rune_id )
    return runeReady
end

function FailStat:IsLogRuneStatistics(info)
    return self.db.logRuneStatistics
end

function FailStat:SetLogRuneStatistics(info, value)
    self.db.logRuneStatistics = value
    self:UpdateLogRuneStatistics()
end

function FailStat:UpdateLogRuneStatistics()    
    if self.db.logRuneStatistics then
        self:RegisterEvents()
    else
        self:UnregisterEvents()
    end
end
